// $Id: DateAxis.java 13825 2011-09-15 15:26:47Z mike $

package org.faceless.graph2;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;
import org.faceless.util.DateParser;

/**
 * A subclass of Axis to be used for plotting date values. Dates are stored
 * internally as doubles, and the two can be converted back and forth using
 * the {@link #toDouble toDouble} and {@link #toDate toDate} methods.
 * For example, to plot a number of dates on a {@link LineSeries} you could
 * use code like the following:
 *
 * <pre class=example>
 * AxesGraph graph = new AxesGraph();
 * LineSeries series = new LineSeries("Dates");
 * series.set(DateAxis.toDouble(date1), value1);
 * series.set(DateAxis.toDouble(date2), value2);
 * graph.setAxis(Axis.BOTTOM, new DateAxis());
 * </pre>
 * (here <i>date1</i> and <i>date2</i> are {@link Date} objects).
 * Bars can also be plotted against a DateAxis - see the
 * {@link GeneralBarSeries} API documentation for more info.
 */
public class DateAxis extends Axis
{
    private static final long MIDPOINT = 631152000;  // 1 Jan 1990
    private static final int SECDAYS = 60*60*24;     // number of seconds in a day
    private static final double MUL = 1000;          // Used to convert doubles to days and vice versa.
                                                     // Needed because of rounding error

    private final DateFormat f;
    private final boolean hasseconds, hashours, hasminutes, hasdays, hasmonths, hasyears;
    private final float mdensity;
    private int barseconds;
    private double stretch;
    private boolean barsatnoon;
    private Calendar c;
    private boolean manualTimezone = false;

    /**
     * Create a new DateAxis with a format of "dd-MMM-yyyy"
     */
    public DateAxis() {
        this(new SimpleDateFormat("dd-MMM-yyyy"));
    }

    /**
     * Create a new DateAxis with the specified format
     */
    public DateAxis(DateFormat format) {
        this(format, DENSITY_NORMAL);
    }

    /**
     * Create a new DateAxis with the specified format and density
     *
     * @param format how each date should be formatted
     * @param density the density of the values on the axis - usually
     * one of {@link Axis#DENSITY_NORMAL}, {@link Axis#DENSITY_SPARSE}
     * or {@link Axis#DENSITY_MINIMAL}, but may be any integer which
     * is roughly the number of intended teeth on the axis.
     */
    public DateAxis(DateFormat format, int density) {
        f = format;
        long l = 1012820522000l; // A random date that has hours, mins, seconds all non-zero.
        String s = f.format(new Date(l));
        hasseconds = !s.equals(f.format(new Date(l+1000l)));
        hasminutes = !s.equals(f.format(new Date(l+60000l)));
        hashours =   !s.equals(f.format(new Date(l+3600000l)));
        hasdays =    !s.equals(f.format(new Date(l+864000000l)));
        hasmonths =  !s.equals(f.format(new Date(l+2419200000l)));      // 28 days in feb
        hasyears  =  !s.equals(f.format(new Date(l+31536000000l)));
        this.mdensity = density/10f;
        if (hasminutes && !hashours && !hasseconds) {
            System.err.println("WARNING: Date format has minutes but not hours or seconds. Did you mean months?");
        }
        c = new GregorianCalendar(hashours ? TimeZone.getDefault() : TimeZone.getTimeZone("UTC"));
    }

    DateFormat getDateFormat() {
        return f;
    }
    boolean hasSeconds() {
        return hasseconds;
    }
    boolean hasMinutes() {
        return hasminutes;
    }
    boolean hasHours() {
        return hashours;
    }
    boolean hasDays() {
        return hasdays;
    }
    boolean hasMonths() {
        return hasmonths;
    }

    public String format(double in) {
        return f.format(toDate(in));
    }

    /**
     * Set the timezone in use for this date axis.
     * @since 2.2.1
     */
    public void setTimeZone(TimeZone tz) {
        c = new GregorianCalendar(tz);
        f.setTimeZone(tz);
        // if we have no time information in the format but have supplied
        // a timezone manually, then we need to respect the timezone and time data
        manualTimezone = true;
    }

    /**
     * When plotting bars against this axis, this method can be used to specify
     * the width of those bars. By default this width is determined
     * automatically based on the format (for example, a format that has a
     * resolution of one day will result in bars one day wide). If the library
     * guesses incorrectly however, this method may be used to hard-code the
     * width. For example, <code>setBarWidth(5*60)</code> would give mean bars
     * drawn against this axis are 5-minutes wide.
     *
     * @param seconds the number of seconds each bar should cover, or 0 to ask
     * the library to guess (the default).
     * @since 2.3
     */
    public void setBarWidth(int seconds) {
        this.barseconds = seconds;
    }

    int getBarWidth() {
        return barseconds;
    }

    /**
     * When plotting a {@link GeneralBarSeries} against a DateAxis, whether
     * to center the bars on the tick, ie. midnight (false) or midway between
     * the tick and the next tick, ie. noon (true). Typically you will want
     * to set this method to true when plotting bars against lines that
     * represent hour values, not just day values. When not plotting Bars,
     * this method has no effect.
     *
     * @param noon whether to center bars plotted against this axis at midday
     */
    public void setBarsAtNoon(boolean noon) {
        this.barsatnoon = noon;
    }

    boolean getBarsAtNoon() {
        return barsatnoon;
    }

    /**
     * <p>
     * Determines whether to "stretch" the ends of the graph to the next useful
     * value. An example of when you would do this is plotting days of a month -
     * although your data might only start on the 3rd of the month, you want the
     * axis to start on the 1st.
     * </p><p>
     * The "stretch" parameter takes a value from 0 to 1, where 0 means "never
     * stretch the axis" and 1 means "always stretch". If necessary, values between
     * 0 and 1 can be specified which will cause the axis to be stretched depending
     * on how far from the next "useful" endpoint the data is.
     * </p>
     * @param stretch when to stretch the axis - 0 for never, 1 for always, or any
     * value in between
     * @since 2.3.1
     */
    public void setStretchEnds(double stretch) {
        this.stretch = stretch!=stretch ? 0 : stretch;
    }

    /**
     * Convert a date to a double, so it can be stored internally.
     * For example, to plot todays date on a Line Graph, you would do
     * something like this:
     * <pre class=example>
     *   LineSeries data = new LineSeries("Dates");
     *   Date today = new Date();
     *   data.set(DateAxis.toDouble(today), 123.45);
     * </pre>
     * @see #toDate
     */
    public static final double toDouble(Date in) {
         return (double)(in.getTime()/MUL)-MIDPOINT;
    }

    /**
     * Convert a double to a Date. The opposite of {@link #toDouble}. Use this
     * to convert a double passed in to the {@link #format} method back to
     * a Date.
     */
    public static final Date toDate(double in) {
        return new Date((long)((in+MIDPOINT)*MUL));
    }

    //-------------------------------------------------------------------------------------------
    // Magic starts here

    public double[] steps(double min, double max) {
        // This is a complicated algorithm. The objective is to return a range of
        // date values to plot on the axis. Values can be VALID or it's subset
        // USEFUL.
        //
        // VALID values are any values that can be formatted without misleading, eg.
        // any whole days when we're plotting days, but not values mid-month if we're
        // plotting only months. So "Jan" is not a valid way to format January 12th,
        // although it is valid for January 1st.
        //
        // USEFUL values are values we'd like to plot to make things neat, eg. 1st of
        // the month or whatever. Useful values are always valid, but not the other
        // way round.
        //
        // Values used by this algorithm are:
        //
        // premin       first USEFUL value less than or equal to min
        // min          the min value
        // postmin      first VALID value greater than or equal to min
        // first        first USEFUL value greater than or equal to min
        // last         first USEFUL value less than or equal to max
        // premax       first VALID value less than or equal to max
        // max          the max value
        // postmax      first USEFUL value greater than or equal to max
        //
        // premin  min  postmin  first      last  premax  max  postmax
        //   |      |      |       |          |     |      |     |
        //   +--<=--+--<=--+--<=---+----------+-<=--+--<=--+-<=--+
        //
        //
        // So, for example, if we're formatting months only from Nov13 to Feb01,
        // premin=Nov01, postmin=first=Dec01, premax=last=Feb01, postmax=Feb01


        // Step 1. Calculate the steps we're going to move in; seconds, minutes,
        // hours or days. Round min/max down to that point.
        //
        long diff = (long)(max-min);    // Difference in seconds between min and max
        int step;
        if (hasseconds && diff<1800) {
            step = 1;
        } else if (hasminutes && diff<43200) {
            step = 60;
        } else if (hashours && diff<SECDAYS*7) {
            step = 3600;
        }
        else if (manualTimezone) {
            // if we have no time information in the format but have supplied
            // a timezone manually, then we need to respect the timezone and time data
            step = 3600;
        }
        else {
            step = SECDAYS;
        }
        min = Math.floor(min/step)*step;  // Round to containing second/minute/hour/day
        max = Math.floor(max/step)*step;

        // Step 2. Iterate from premin to postmax, defining the values as we go.
        // Add values from "first" to "last" to the list.
        //
        List list = new ArrayList();
        double premin=min, postmin, first, last, premax, postmax=max;
        postmin = first = last = premax = Double.NaN;
        while (!isUseful(toDate(premin), diff, step)) {
            premin -= step;     // Step backwards until we're useful.
        }
        double val = premin;
        while (val<=max) {
            Date date = toDate(val);
            if (isValid(date)) {
                if (val>=min) {
                    if (postmin!=postmin) postmin = val;
                    premax = val;
                    if (isUseful(date, diff, step)) {
                        if (first!=first) first = val;
                        last = val;
                        list.add(new Double(val));
                    }
                }
            }
            val += step;
        }
        while (!isUseful(toDate(postmax), diff, step)) {
            postmax += step;
        }

        // Step 3. Modify the end points if necessary. We do this by either:
        // a) stretching the ends to premin/postmax - if stretch > 0
        // b) adding min/max to the list - if they're valid and not too close
        //    to first/last
        // c) replacing first/last with min/max - if they're valid and are too
        //    close to first/last (ie they'd cramp the graph)
        // d) not adding min/max to the list - if they're not valid.

        if (list.size()<2) {
            list.clear();
            list.add(new Double(premin));
            if (premin!=postmax) list.add(new Double(postmax));
        } else {
            if ((postmin-min)/(postmin-premin) < stretch) {
                list.add(0, new Double(premin));
            } else if (first!=min && isValid(toDate(min))) {
                if ((first-min)/(max-min) < 0.05/mdensity) {
                    list.set(0, new Double(min));
                } else {
                    list.add(0, new Double(min));
                }
            }
            if ((max-premax)/(postmax-premax) < stretch) {
                list.add(new Double(postmax));
            } else if (last!=max && isValid(toDate(max))) {
                if ((max-last)/(max-min) < 0.05/mdensity) {
                    list.set(list.size()-1, new Double(max));
                } else {
                    list.add(new Double(max));
                }
            }
        }

        // We're done, return the list as a double[]

        double[] out = new double[list.size()];
        for (int i=0;i<out.length;i++) {
            out[i] = ((Double)list.get(i)).doubleValue();
        }
        return out;
    }

    /**
     * Return whether this date is "valid"
     */
    private boolean isValid(Date date) {
        c.setTime(date);
        if (!hasminutes && !manualTimezone && c.get(c.MINUTE)!=0) return false;
        if (!hashours && !manualTimezone && c.get(c.HOUR_OF_DAY)!=0) return false;
        if (!hashours && !manualTimezone && !hasdays && c.get(c.DAY_OF_MONTH)!=1) return false;
        if (!hashours && !manualTimezone && !hasmonths && c.get(c.MONTH)!=0) return false;
        return true;
    }

    /**
     * General method for parsing strings to dates as best we can
     *
     * @param value Date represented as a string
     * @return Date object for string passed in
     * @throws Exception
     */
    protected final Date parse(String value) {
        Date date = null;
        try {
            date = f.parse(value);
        } catch (Exception e) {
            date = DateParser.parse(value);
            if (date == null) {
                throw new IllegalArgumentException("Cannot parse date \""+value+"\"");
            }
        }
        return date;
    }

    /**
     * Return whether this date is "useful"
     */
    private boolean isUseful(Date date, double diff, int step) {
        boolean useful = false;
        if (isValid(date)) {
            if (hasseconds && diff<=15*mdensity) {              // 15 seconds - every seconds, tot<=15
                useful = true;
            } else if (hasseconds && diff<=60*mdensity) {       // 1 minutes - every 5 seconds, tot<=12
                useful = c.get(c.SECOND)%5==0;
            } else if (hasseconds && diff<=180*mdensity) {      // 3 minutes - every 15 seconds, tot<=12
                useful = c.get(c.SECOND)%15==0;
            } else if (hasseconds && diff<=300*mdensity) {      // 5 minutes - every 30 seconds, tot<=10
                useful = c.get(c.SECOND)%30==0;
            } else if (hasminutes && diff<=900*mdensity) {      // 15 minutes - every minute, tot<=15
                useful = c.get(c.SECOND)==0;
            } else if (hasminutes && diff<=3600*mdensity) {       // 1 hour - every 5 minutes, tot<=12
                useful = c.get(c.MINUTE)%5==0 && c.get(c.SECOND)==0;
            } else if (hasminutes && diff<=7200*mdensity) {       // 2 hours - every 10 minutes, tot<=12
                useful = c.get(c.MINUTE)%10==0 && c.get(c.SECOND)==0;
            } else if (hasminutes && diff<=10800*mdensity) {      // 3 hours - every 15 minutes, tot<=12
                useful = c.get(c.MINUTE)%15==0 && c.get(c.SECOND)==0;
            } else if (hasminutes && diff<=21600*mdensity) {      // 6 hours - every 30 minutes, tot<=12
                useful = c.get(c.MINUTE)%30==0 && c.get(c.SECOND)==0;
            } else if (hashours && diff<=SECDAYS*mdensity) {      // 24 hours - every 2 hours, tot<=12
                useful = c.get(c.HOUR_OF_DAY)%2==0 && c.get(c.MINUTE)==0;
            } else if (hashours && diff<4*SECDAYS*mdensity) {     // 96 hours (4 days) - every 6 hours, tot<15
                useful = c.get(c.HOUR_OF_DAY)%6==0 && c.get(c.MINUTE)==0;
            } else if (hashours && diff<=6*SECDAYS*mdensity) {    // 144 hours (6 days) - ev. 12 hours, tot<=12
                useful = c.get(c.HOUR_OF_DAY)%12==0 && c.get(c.MINUTE)==0;
            } else if (hasdays && diff<15*SECDAYS*mdensity) {     // 2 weeks - every day, tot<=14
                useful = step>3600 || c.get(c.HOUR_OF_DAY)==0;
            } else if (hasdays && diff<32*SECDAYS*mdensity) {     // 1 month - every 2 days, tot<=15
                useful = (c.get(c.DAY_OF_MONTH)&1)==1 && c.get(c.HOUR_OF_DAY)==0;
            } else if (hasdays && diff<100*SECDAYS*mdensity) {    // 3 months - every sunday, tot<=13
                useful = c.get(c.DAY_OF_WEEK)==1 && c.get(c.HOUR_OF_DAY)==0;
            } else if (hasdays && diff<200*SECDAYS*mdensity) {    // 6 months - every 1st and 15th, tot<=12
                useful = (c.get(c.DAY_OF_MONTH)==1 || c.get(c.DAY_OF_MONTH)==15) && c.get(c.HOUR_OF_DAY)==0;
            } else if (hasmonths && diff<367*SECDAYS*mdensity) {  // 1 year - every 1st, tot<=12
                useful = c.get(c.DAY_OF_MONTH)==1 && c.get(c.HOUR_OF_DAY)==0;
            } else if (hasmonths && diff<730*SECDAYS*mdensity) {  // 2 years - 1st of every 2nd month, tot<=12
                useful = c.get(c.DAY_OF_MONTH)==1 && c.get(c.MONTH)%2==0 && c.get(c.HOUR_OF_DAY)==0;
            } else if (hasmonths && diff<2190*SECDAYS*mdensity) { // 3 years - 1st of every quarter, tot<=12
                useful = c.get(c.DAY_OF_MONTH)==1 && c.get(c.MONTH)%3==0 && c.get(c.HOUR_OF_DAY)==0;
            } else if (hasmonths && diff<4400*SECDAYS*mdensity) { // 6 years - 1st of jan/jul, tot<=12
                useful = c.get(c.DAY_OF_MONTH)==1 && c.get(c.MONTH)%6==0 && c.get(c.HOUR_OF_DAY)==0;
            } else if (diff<5478*SECDAYS*mdensity) {              // 15 years - 1st of jan, tot<=15
                useful = c.get(c.DAY_OF_MONTH)==1 && c.get(c.MONTH)==0 && c.get(c.HOUR_OF_DAY)==0;
            } else {                                              // Over 15 years - Jan 1st every n years
                int y = (int)(diff/(5478*SECDAYS*mdensity))+1;
                useful = c.get(c.DAY_OF_MONTH)==1 && c.get(c.MONTH)==0 && (c.get(c.YEAR)%y)==0 && c.get(c.HOUR_OF_DAY)==0;
            }
        }
        return useful;
    }
}
